import argparse
from .info import RequiredModules

def parse_arguments():
    parser = argparse.ArgumentParser(description="Get required modules safe versions!")
    parser.add_argument('project_name', help='target project name.       ex) mmdet3d')
    parser.add_argument('project_version', help='target project version. ex) 0.17.0')
    parser.add_argument('req_path', help='target requirement.txt path.   ex) ./new_reqirement.txt')
    parser.add_argument('--export-path', help='export pretty requirement.txt to {export_path}')

    return parser.parse_args()

def main():
    args = parse_arguments()

    req_modules = RequiredModules.get_requirement_text(
        path=args.req_path,
        target_project=args.project_name,
        target_project_version=args.project_version
    )

    if args.export_path is not None:
        req_modules.export(args.export_path)
        outp = f"Export pretty requirement.txt to '{args.export_path}'  "
        print("\n[SUCCESS]")
        print("="*len(outp))
        print(outp)
        print("="*len(outp))
    else:
        print("\n[OUTPUT]")
        info_width = max([len(info) for info in req_modules.__repr__().split("\n")]) + 2
        print("="*info_width)
        print(req_modules)
        print("="*info_width)



import requests
from bs4 import BeautifulSoup

from datetime import datetime

from typing import Optional

class ReleaseHistory:
    def __init__(self, project_name:str, release_history:dict):
        self.project_name = project_name
        self.version_to_datetime = release_history

    def get_closest_version_by_datetime(self, target_datetime: datetime) -> str:
        target_timestamp = target_datetime.timestamp()

        minimum_diff = -1
        closest_version = None
        for k,v in self.version_to_datetime.items():
            diff = target_timestamp - v["datetime"].timestamp()
            if diff < 0:
                continue
            if (closest_version is None) or (diff < minimum_diff):
                minimum_diff = diff
                closest_version = k
                break
        
        # See https://github.com/uinone/package-version-seeker/issues/1
        if closest_version is None:
            versions = list(self.version_to_datetime.keys())
            closest_version = versions[-1] # Get oldest version

        return f"=={closest_version}"
    
    def get_datetime_by_version(self, version:str) -> Optional[datetime]:
        try:
            dt = self.version_to_datetime[version]
            return dt
        except KeyError:
            print(f"version '{version}' does not exist in {self.project_name}'s release history\nCheck 'https://pypi.org/project/{self.project_name}/#history'")
            return None

    def __repr__(self) -> str:
        return f"[Project] {self.project_name}"

    @classmethod
    def get_project_release_history(cls, project_name:str):
        release_history_url = f"https://pypi.org/project/{project_name}/#history"

        release_history_response = requests.get(release_history_url)
        soup = BeautifulSoup(release_history_response.text, "html.parser")

        time_line = soup.find(name="div", attrs="release-timeline")
        
        releases = time_line.findAll(name="div", attrs="release")

        release_history = {}
        for release in releases:
            version = release.find(name="p", attrs="release__version").text.strip().split("\n")[0]
            dt = release.find(name="time").attrs["datetime"].split("+")[0]
            release_history[version] = {"datetime": datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S")}
        
        return ReleaseHistory(
            project_name=project_name,
            release_history=release_history
        )


from .history import ReleaseHistory

from typing import Optional

class RequiredModules:
    def __init__(self, infos:dict):
        self.infos = infos

    def __repr__(self) -> str:
        repr_text = ""

        for k, v in self.infos.items():
            if v is None:
                repr_text += k
                repr_text += "\n"
            else:
                repr_text += k + v
                repr_text += "\n"
        
        return repr_text[:-1]
    
    def _fill_empty_versions(self, target_project:str, target_project_version:str):
        module_with_empty_version = {}

        # Aggregating
        for k, v in self.infos.items():
            if v is None:
                module_with_empty_version[k] = None
        
        target_project_release_hist = ReleaseHistory.get_project_release_history(target_project)
        target_datetime = target_project_release_hist.get_datetime_by_version(target_project_version)

        if target_datetime is None:
            return None

        target_datetime = target_datetime["datetime"]

        for module_name in module_with_empty_version.keys():
            release_hist = ReleaseHistory.get_project_release_history(module_name)
            closest_version = release_hist.get_closest_version_by_datetime(target_datetime)
            module_with_empty_version[module_name] = closest_version
        
        self.infos.update(module_with_empty_version)

        return self
    
    def export(self, path:str):
        with open(path, "w") as f:
            f.write(self.__repr__())


    @classmethod
    def get_requirement_text(cls, path:str, target_project:str, target_project_version:str) -> Optional[ReleaseHistory]:
        required_module_infos = {}
        try:
            with open(path, "r") as f:
                required_module_names = [module_name for module_name in f.read().split("\n") if not(module_name.startswith("#")) and len(module_name) > 0]
                
                for module_name in required_module_names:
                    split_prefix_candidate = [module_name.find("<"), module_name.find("="), module_name.find(">")]
                    split_prefix_candidate.sort()

                    postive_index = -1
                    for cand in split_prefix_candidate:
                        if cand > 0:
                            postive_index = cand
                            break

                    if postive_index == -1:
                        required_module_infos[module_name] = None
                    else:
                        required_module_infos[module_name[:postive_index]] = module_name[postive_index:]
            
            return RequiredModules(infos=required_module_infos)._fill_empty_versions(target_project, target_project_version)
        except Exception as e:
            print(f"[FAILED]\n{e}")


from setuptools import setup, find_packages

def readme():
    with open('README.md', encoding='utf-8') as f:
        content = f.read()
    return content

setup(
    name='pakage-version-seeker',
    version='1.0.1',
    description='Find unspecified pakages verison!!',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='uinone',
    author_email='tizm423@gmail.com',
    url='https://github.com/uinone/package-version-seeker',
    packages=find_packages(),
    include_package_data=True,
    keywords=[],
    python_requires='>=3.8',
    package_data={},
    entry_points={'console_scripts': ["pvs=src:main"]},
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    install_requires=['requests==2.32.3', 'beautifulsoup4==4.12.3',],
    license='MIT',
)